exec("./sounds.cs");

function music()
{
	exec("./server.cs");
}

function parseFileABC(%file, %primary, %secondary, %transpose)
{
	if(isObject($ABC::SO))
	{
		$ABC::SO.delete();
	}
	
	%fo = new FileObject();
	%fo.openForRead(%file);

	return __parseFileABCSchedule(%text, %file, %primary, %secondary, %transpose, %fo);
}

function __parseFileABCSchedule(%text, %file, %primary, %secondary, %transpose, %fo, %count)
{
	cancel($ABC::ReadSchedule);
	if(!isObject(%fo))
	{
		return;
	}
	if(%fo.isEOF())
	{
		%fo.close();
		%fo.delete();
		return parseABC(%text, %file, %primary, %secondary, %transpose);
	}
	else
	{
		bottomPrintAll("<font:impact:24>\c6Reading line" SPC %count, 1);
		%text = %text @ (%text $= "" ? "" : "\n") @  %fo.readLine();
		
		$ABC::ReadSchedule = schedule(18, 0, __parseFileABCSchedule, %text, %file, %primary, %secondary, %transpose, %fo, %count++);
	}
}

function parseABC(%text, %fileName, %primary, %secondary, %transpose)
{
	%count = getLineCount(%text);
	%unitLength = 1;

	%accPending = "";
	%accNum = 0;
	
	if(isObject($ABC::SO))
	{
		$ABC::SO.delete();
	}

	%obj = new ScriptObject()
	{
		class = ABCData;
		size = 0;
		bpm = 120;

		primary = %primary;
		secondary = %secondary;
		transpose = %transpose;
	};
	
	$ABC::SO = %obj;

	for (%j = 0; %j < %count; %j++)
	{
		%line = getLine(%text, %j);
		%size = strLen(%line);

		// Check for comment lines
		if (getSubStr(%line, 0, 1) $= "%")
			continue;

		// Check for meta lines
		if (getSubStr(%line, 1, 1) $= ":")
		{
			%key = getSubStr(%line, 0, 1);
			%val = getSubStr(%line, 2, %size);

			%obj.meta[%key] = %val;
			echo(%key SPC "=" SPC %val);

			if (%key $= "L")
			{
				%right = nextToken(%val, "left", "/");
				%unitLength = %left / %right;
			}

			if (%key $= "Q" && strPos(%val, "=") == -1)
				%obj.bpm = mFloor(%val);

			// EXTREMELY IMPORANT
			// PROPERLY ADJUST TO KEY SCALE
			// CURRENTLY ASSUMES C SCALE ALWAYS
			if (%key $= "K")
			{
				// ???
			}

			continue;
		}

		%chordSize = "";
		%chordLength = "";

		for (%i = 0; %i < %size + 1; %i++)
		{
			%char = getSubStr(%line, %i, 1);

			// Reset ALL THE THINGS (only accidentals and chords)
			// Empty char is a hack to reset on EOF
			if (%char $= "|" || %char $= "")
			{
				%data = "";

				for (%k = 0; %k < %accNum; %k++)
				{
					%accVal[%accKey[%k]] = "";
					%accKey[%k] = "";
				}

				%accNum = 0;

				if (%chordSize)
				{
					for (%k = 0; %k < %chordSize; %k++)
						%chordKey[%k] = "";

					%chordSize = "";
					%chordLength = "";
				}

				continue;
			}

			// Ignore these for now, need to parse them eventually
			if (strPos(" :", %char) != -1)
				continue;

			%data = %data @ %char;

			// Trying to parse chords
			if (%char $= "[")
			{
				if (%chordSize $= "")
					%chordSize = 0;

				continue;
			}

			if (%char $= "]")
			{
				if (%chordSize)
				{
					for (%k = 0; %k < %chordSize; %k++)
					{
						%obj.key[%obj.size] = %obj.key[%obj.size] @ (%k ? " " : "") @ %chordKey[%k];
						%chordKey[%k] = "";
					}

					%obj.length[%obj.size] = %chordLength;
					%obj.data[%obj.size] = %data;
					%obj.time += %chordLength;
					%obj.size++;

					%chordLength = "";
					%chordSize = "";
					%data = "";
				}

				continue;
			}

			// Read accidentals
			%index = strPos("_=^", %char);

			if (%index != -1)
			{
				%accPending += %index - 1;
				continue;
			}

			// At this point, %char has to be a note or rest symbol
			%key = charToBaseKey(%char);

			if (%key $= "")
			{
				// Is this really the right thing to do?
				// Shouldn't we just crash() or something?
				//echo("Error: Invalid character '" @ %char @ "'");
				continue;
			}

			if (%key != -1) // Only do this for notes
			{
				// Base transpose
				%key += 60;

				for (_; %i < %size; %i++)
				{
					%next = getSubStr(%line, %i + 1, 1);

					if (%next $= "'")
					{
						%key += 12;
						%data = %data @ %next;
					}
					else if (%next $= ",")
					{
						%key -= 12;
						%data = %data @ %next;
					}
					else
						break;
				}

				// Now we have the (mostly) final key index,
				// assign accidental changes to it
				if (%accPending !$= "")
				{
					if (%accVal[%key] $= "")
					{
						%accKey[%accNum] = %key;
						%accNum++;
					}

					%accVal[%key] = %accPending;
					%accPending = "";
				}

				// .. and apply the accidental
				%key += %accVal[%key];
			}

			// Now we need the time signature if there is one
			// Do this for rests too, obviously

			%left = 1;
			%right = 1;

			for (%start = %i; %i < %size; %i++)
			{
				%next = getSubStr(%line, %i + 1, 1);

				if (%next !$= mFloor(%next))
					break;
			}

			if (%i != %start)
			{
				%left = getSubStr(%line, %start + 1, %i - %start);
				%data = %data @ %left;
			}

			if (getSubStr(%line, %i + 1, 1) $= "/")
			{
				%i++;

				for (%start = %i; %i < %size; %i++)
				{
					%next = getSubStr(%line, %i + 1, 1);

					if (%next !$= mFloor(%next))
					{
						break;
					}
				}

				if (%i != %start)
				{
					%right = getSubStr(%line, %start + 1, %i - %start);
					%data = %data @ "/" @ %right;
				}
			}

			// That was a lot..
			%length = %left / %right;

			// Not sure if we should be doing this
			%length *= %unitLength;

			// I think we're done! .. mostly
			// need to parse a ton of other stuff at some point too

			// If we're in a chord, add the new key to the list
			if (%chordSize !$= "")
			{
				%chordKey[%chordSize] = %key;

				if (%chordLength $= "" || %length < %chordLength)
					%chordLength = %length;

				%chordSize++;
			}
			else
			{
				// Just add to the %obj like normal
				%obj.key[%obj.size] = %key;
				%obj.length[%obj.size] = %length;
				%obj.data[%obj.size] = %data;
				%obj.time += %length;
				%data = "";

				%obj.size++;
			}
		}
	}

	%obj.play(0);
	return %obj;
}

function charToBaseKey(%char)
{
	switch$ (%char)
	{
		case "C": %key = 0;
		case "D": %key = 2;
		case "E": %key = 4;
		case "F": %key = 5;
		case "G": %key = 7;
		case "A": %key = 9;
		case "B": %key = 11;
		case "Z": return -1;
		case "X": return -1;

		default: return "";
	}

	if (strStr(%char, strLwr(%char)) == 0)
		%key += 12;

	return %key;
}

function ABCData::play(%this, %i, %time)
{
	cancel(%this.play);

	if (%i $= "")
		%i = 0;

	if (%i >= %this.size)
		return;

	%speed = (%this.bpm / 60) * (1 / 4);
	//%speed = 0.6;
	%length = %this.length[%i] / %speed;
	%duration = %this.time / %speed;

	%step = 0.075;
	%count = getWordCount(%this.key[%i]);

	for (%j = 0; %j < %count; %j++)
	{
		%key = getWord(%this.key[%i], %j);

		if (%key != -1)
		{
			%key += %this.transpose;

			if (%this.secondary $= "" || %key >= $InstrumentLow[%this.primary])
				playInstrument(%this.primary, %key);
			else
				playInstrument(%this.secondary, %key);
		}
	}

	%a = getTimeString(mFloor(%time));
	%b = getTimeString(mFloor(%duration));

	bottomPrintAll("<font:impact:24>\c6" @ %this.data[%i] @ "\n<font:impact:24>\c6" @ %i + 1 @ "/" @ %this.size @ " - " @ %a @ "/" @ %b @ "\n<font:impact:24>\c6" @ %this.meta["T"], %length + 0.25);
	//announce(%length);
	%this.play = %this.schedule(%length * 1000, play, %i + 1, %time + %length);
}

function playInstrument(%type, %note)
{
	//announce(%type SPC %note);
	%range = $InstrumentHigh[%type] - $InstrumentLow[%type];
	%scale = 1;

	if (%note < $InstrumentLow[%type])
	{
		%scale = 0.5;
		%note += %range;
	}
	else if (%note > $InstrumentHigh[%type])
	{
		%scale = 2;
		%note -= %range;
	}

	if (%note < $InstrumentLow[%type] || %note > $InstrumentHigh[%type])
		return;

	%sound = $InstrumentSound[%type, %note];

	if (!isObject(%sound))
		return;

	%current = getTimeScale();

	if (%current != %scale)
		commandToAll('TimeScale', %scale);

	serverPlay2D(%sound);

	if (%current != %scale)
		commandToAll('TimeScale', %current);
}